package packageForLT06;

import java.awt.Color;
import java.awt.Point;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.util.List;
import java.util.Random;
import java.util.Vector;

/**
 * <p>
 * The class encapsulates a data model that consists of a set of dots. The dots
 * are uniquely identified by an integer index. The first dot's identifier is 0,
 * the second's is 1, and so on until the final dot.
 * </p>
 * 
 * <p>
 * Each dot is understood to have a diameter (of a certain number of units,
 * where the actual physical size of the unit is left unspecified and is
 * abstracted away). Each dot is understood to have a location, which is
 * specified as an x, y coordinate in a two-dimensional plane that has its
 * origin in the upper left corner. A two-dimensional plane of 200x600 units is
 * assumed. Each dot is understood to be associated with a color.
 * </p>
 * 
 * <p>
 * Each dot is randomly assigned a position, diameter and color. The positions
 * are chosen randomly from among the visible positions. The diameters are
 * chosen randomly from among the possible values in the interval [50,100]. The
 * colors are chosen randomly from the colors red, blue, and green.
 * </p>
 * 
 * <p>
 * This model is capable of having listeners installed upon it. Anytime this
 * model changes, the listener's <tt>changed()</tt> method will be invoked.
 * </p>
 * 
 * @author mb
 * 
 */
public class DotModel {
	private List<Ellipse2D> dots;
	private List<Color> colours;
	private int DIAMETER_LOWER_BOUND = 50;
	private int DIAMETER_UPPER_BOUND = 100;
	private int MAX_X = 200;
	private int MAX_Y = 600;
	private int HEADER = 30;

	private int MIN_DOTS = 2;
	private int MAX_DOTS = 20;

	private Color[] palette = { Color.RED, Color.BLUE, Color.GREEN };
	private int NUDGE = 4;

	private Random random = new Random();

	private List<DotModelListener> listeners;

	/**
	 */

	/**
	 * Creates a default version of this data model with the specified number of
	 * dots. Each dot is randomly assigned a position, diameter and color. The
	 * positions are chosen randomly from among the visible positions. The
	 * diameters are chosen randomly from among the possible values in the
	 * interval [50,100]. The colors are chosen randomly from the colors red,
	 * blue, and green.
	 * 
	 * 
	 * @param numberOfDots
	 *            the number of dots this model should contain. Must be a number
	 *            between 2 and 10.
	 * @throws RuntimeException
	 *             if the number of dots is invalid
	 */
	public DotModel(int numberOfDots) throws RuntimeException {
		if (numberOfDots > MAX_DOTS || numberOfDots < MIN_DOTS)
			throw new RuntimeException(
					"Invalid number of dots specified for this model");
		dots = new Vector<Ellipse2D>();
		colours = new Vector<Color>();
		for (int i = 0; i < numberOfDots; i++) {
			int diameter = random.nextInt(DIAMETER_UPPER_BOUND
					- DIAMETER_LOWER_BOUND)
					+ DIAMETER_LOWER_BOUND;
			int colorIdx = random.nextInt(palette.length);
			int xPos = random.nextInt(MAX_X - diameter);
			int yPos = random.nextInt(MAX_Y - diameter - HEADER);
			Ellipse2D dot = new Ellipse2D.Double(xPos, yPos, diameter, diameter);
			colours.add(palette[colorIdx]);
			dots.add(dot);
			System.out.println("Created Dot: (x,y) diameter colour: "
					+ dot.getX() + "\t" + dot.getY() + "\t" + dot.getWidth()
					+ "\t" + palette[colorIdx]);
		}
		listeners = new Vector<DotModelListener>();
	}

	/**
	 * Causes the indicated dot to be nudged to the right by a small amount.
	 * 
	 * If any listeners are installed on this model, they will be informed that
	 * the model has changed.
	 * 
	 */
	public void nudgeToRight(int dotIdentifier) {
		Ellipse2D theDot = dots.get(dotIdentifier);
		theDot = new Ellipse2D.Double(theDot.getX() + NUDGE, theDot.getY(),
				theDot.getWidth(), theDot.getHeight());
		dots.remove(dotIdentifier);
		dots.add(dotIdentifier, theDot);
		notifyModelHasChanged();
	}

	/**
	 * Causes the indicated dot to have a new diameter as specified by the
	 * passed value. The location point of the specified dot will not be
	 * changed.
	 * 
	 * If any listeners are installed on this model, they will be informed that
	 * the model has changed.
	 */
	public void changeDiameter(int dotIdentifier, int newDiameter) {
		Ellipse2D theDot = dots.get(dotIdentifier);
		theDot = new Ellipse2D.Double(theDot.getX(), theDot.getY(),
				newDiameter, newDiameter);
		dots.remove(dotIdentifier);
		dots.add(dotIdentifier, theDot);
		notifyModelHasChanged();
	}

	/**
	 * Gets the maximum horizontal dimension.
	 * 
	 * @return
	 */
	public int getMaxX() {
		return MAX_X;
	}

	/**
	 * Gets the maximum vertical dimension.
	 * 
	 * @return
	 */
	public int getMaxY() {
		return MAX_Y;
	}

	/**
	 * Installs the passed listener on this model.
	 * 
	 * @param listener
	 */
	public void addListener(DotModelListener listener) {
		listeners.add(listener);
	}

	private void notifyModelHasChanged() {
		for (DotModelListener listener : listeners) {
			listener.changed();
		}
	}

	public int getNumberOfDots() {
		return dots.size();
	}

	/**
	 * Gets the colour that corresponds to the specified dot.
	 * 
	 * @param dotIdentifier
	 *            the identifier of the dot
	 * @throws RuntimeException
	 *             if the identifier for the dots is invalid
	 */
	public Color getColour(int dotIdentifier) throws RuntimeException {
		if (dotIdentifier > dots.size() - 1 || dotIdentifier < 0)
			throw new RuntimeException("Invalid dot identifier specified.");
		return colours.get(dotIdentifier);
	}

	/**
	 * Sets the colour of the specified dot to the passed color.
	 * 
	 * @param dotIdentifier
	 *            the identifier of the dot
	 * @throws RuntimeException
	 *             if the identifier for the dots is invalid
	 */
	public void setColour(int dotIdentifier, Color newColor)
			throws RuntimeException {
		if (dotIdentifier > dots.size() - 1 || dotIdentifier < 0)
			throw new RuntimeException("Invalid dot identifier specified.");
		colours.set(dotIdentifier, newColor);
		notifyModelHasChanged();
	}

	/**
	 * Gets the diameter of the specified dot.
	 * 
	 * @param dotIdentifier
	 *            the identifier of the dot
	 * @throws RuntimeException
	 *             if the identifier for the dots is invalid
	 */
	public int getDiameter(int dotIdentifier) {
		if (dotIdentifier > dots.size() - 1 || dotIdentifier < 0)
			throw new RuntimeException("Invalid dot identifier specified.");
		return (int) dots.get(dotIdentifier).getWidth();
	}

	/**
	 * Gets the dot that corresponds to the specified dot Identifier.
	 * 
	 * @param dotIdentifier
	 *            the identifier of the dot
	 * @throws RuntimeException
	 *             if the identifier for the dots is invalid
	 */
	public Ellipse2D getDot(int dotIdentifier) throws RuntimeException {
		if (dotIdentifier > dots.size() - 1 || dotIdentifier < 0)
			throw new RuntimeException("Invalid dot identifier specified.");
		return dots.get(dotIdentifier);
	}

	/**
	 * Gets the dot that is closest to the specified point, as measured to the
	 * dot's center point.
	 * 
	 * @param point
	 *            the Point
	 */
	public int getClosestDotToPoint(Point point) {
		double minDist = getDistance(new Point(0, 0), new Point(MAX_X, MAX_Y));
		int index = 0;
		for (int i = 0; i < this.getNumberOfDots(); i++) {
			Point currDotCenter = new Point((int) (Math.round(getDot(i)
					.getCenterX())), (int) (Math.round(getDot(i).getCenterY())));
			double currentDist = getDistance(currDotCenter, point);
			if (currentDist < minDist) {
				index = i;
				minDist = currentDist;
			}
		}
		return index;
	}

	private double getDistance(Point point1, Point point2) {
		return Math.sqrt(Math.pow(point1.x - point2.x, 2)
				+ Math.pow(point1.y - point2.y, 2));
	}

}
